/*!\file MisfitAnnual.cpp
 * \brief: MisfitAnnual object
 */

/*Headers:*/
/*{{{*/
#ifdef HAVE_CONFIG_H
   #include <config.h>
#else
#error "Cannot compile with HAVE_CONFIG_H symbol! run configure first!"
#endif

#include <float.h>    /*  DBL_EPSILON  */
#include "./classes.h"
#include "./ExternalResults/ExternalResult.h"
#include "./ExternalResults/Results.h"
#include "../datastructures/datastructures.h"
#include "./Elements/Element.h"
#include "./Elements/Elements.h"
#include "./FemModel.h"
#include "../modules/SurfaceAreax/SurfaceAreax.h"
#include "../classes/Params/Parameters.h"
#include "../classes/gauss/Gauss.h"
/*}}}*/

/*MisfitAnnual constructors, destructors :*/
MisfitAnnual::MisfitAnnual(){/*{{{*/

	this->definitionenum = -1;
	this->name = NULL;
	this->model_enum = UNDEF;
	this->observation_enum = UNDEF;
	this->weights_enum = UNDEF;
	this->timeinterpolation=NULL;
	this->local=1;
	this->misfit=0;
	this->annual_dt=0;
	this->initialized=0;
	this->lock=0;

}
/*}}}*/
MisfitAnnual::MisfitAnnual(char* in_name, int in_definitionenum, int in_model_enum, int in_observation_enum, char* in_timeinterpolation, int in_local, int in_weights_enum){/*{{{*/

	this->definitionenum=in_definitionenum;

	this->name		= xNew<char>(strlen(in_name)+1);
	xMemCpy<char>(this->name,in_name,strlen(in_name)+1);

	this->timeinterpolation = xNew<char>(strlen(in_timeinterpolation)+1);
	xMemCpy<char>(this->timeinterpolation,in_timeinterpolation,strlen(in_timeinterpolation)+1);

	this->model_enum=in_model_enum;
	this->observation_enum=in_observation_enum;
	this->weights_enum=in_weights_enum;
	this->local=in_local;

	this->misfit=0;
	this->annual_dt=0;
	this->initialized=0;
	this->lock=0;
}
/*}}}*/
MisfitAnnual::~MisfitAnnual(){/*{{{*/
	if(this->name)xDelete(this->name);
	if(this->timeinterpolation)xDelete(this->timeinterpolation);
	this->misfit=0;
	this->annual_dt=0;
	this->initialized=0;
	this->lock=0;
}
/*}}}*/
/*Object virtual function resolutoin: */
Object* MisfitAnnual::copy() {/*{{{*/
	MisfitAnnual* mf = new MisfitAnnual(this->name,this->definitionenum, this->model_enum,this->observation_enum,this->timeinterpolation,this->local,this->weights_enum);
	mf->misfit=this->misfit;
	mf->annual_dt=this->annual_dt;
	mf->lock=this->lock;
	return (Object*) mf;
}
/*}}}*/
void MisfitAnnual::DeepEcho(void){/*{{{*/
	this->Echo();
}
/*}}}*/
void MisfitAnnual::Echo(void){/*{{{*/
	_printf_(" MisfitAnnual: " << name << " " << this->definitionenum << "\n");
	_printf_("    model_enum: " << model_enum << " " << EnumToStringx(model_enum) << "\n");
	_printf_("    observation_enum: " << observation_enum << " " << EnumToStringx(observation_enum) << "\n");
	_printf_("    weights_enum: " << weights_enum << " " << EnumToStringx(weights_enum) << "\n");
	_printf_("    timeinterpolation: " << timeinterpolation << "\n");
	_printf_("    local: " << local << "\n");
}
/*}}}*/
int MisfitAnnual::Id(void){/*{{{*/
	return -1;
}
/*}}}*/
void MisfitAnnual::Marshall(MarshallHandle* marshallhandle){/*{{{*/
	_error_("not implemented yet!");
}
/*}}}*/
int MisfitAnnual::ObjectEnum(void){/*{{{*/
	return MisfitannualEnum;
}
/*}}}*/
/*Definition virtual function resolutoin: */
int MisfitAnnual::DefinitionEnum(){/*{{{*/
	return this->definitionenum;
}
/*}}}*/
char* MisfitAnnual::Name(){/*{{{*/
	char* name2=xNew<char>(strlen(this->name)+1);
	xMemCpy(name2,this->name,strlen(this->name)+1);

	return name2;
}
/*}}}*/
IssmDouble MisfitAnnual::Response(FemModel* femmodel){/*{{{*/

    /*diverse: */
    IssmDouble time,starttime,finaltime;
    IssmDouble dt;

    /*recover time parameters: */
    femmodel->parameters->FindParam(&starttime,TimesteppingStartTimeEnum);
    femmodel->parameters->FindParam(&finaltime,TimesteppingFinalTimeEnum);
    femmodel->parameters->FindParam(&time,TimeEnum);
    femmodel->parameters->FindParam(&dt,TimesteppingTimeStepEnum);

    if (this->local==1){ /*area integration using elements: {{{*/
		/*Do we lock? i.e. are we at final_time? :*/  
		_printf0_(" MisfitAnnual::133  lock: " << this->lock << "\n");
		if(this->lock>=1) {
			IssmDouble misfit_t=0.;
			IssmDouble all_misfit_t=0.;
			IssmDouble area_t=0.;
			IssmDouble all_area_t=0.;
			IssmDouble area_e = 0.;


			/*parameters: */
			IssmDouble starttime,finaltime,dt,yts;
			bool       iscontrol,isautodiff;
			int        timestepping;
			int        output_frequency,checkpoint_frequency;
			int        amr_frequency;
			char     **requested_outputs = NULL;
			IssmDouble* element_accum_values = xNewZeroInit<IssmDouble>(3);

			/*intermediary: */
			int        step;
			IssmDouble time;
			IssmDouble annual_dt;

			/*first, figure out if there was a check point, if so, do a reset of the FemModel* femmodel structure. */
			// femmodel->parameters->FindParam(&checkpoint_frequency,SettingsCheckpointFrequencyEnum);
			// if(checkpoint_frequency) femmodel->Restart();

			/*then recover parameters common to all solutions*/
			// femmodel->parameters->FindParam(&step,StepEnum);
			// femmodel->parameters->FindParam(&time,TimeEnum);
			femmodel->parameters->FindParam(&starttime,TimesteppingStartTimeEnum);
			femmodel->parameters->FindParam(&finaltime,TimesteppingFinalTimeEnum);
			femmodel->parameters->FindParam(&yts,ConstantsYtsEnum);
			femmodel->parameters->FindParam(&output_frequency,SettingsOutputFrequencyEnum);
			femmodel->parameters->FindParam(&timestepping,TimesteppingTypeEnum);

			/*initialize:  */
			step = 0;
			time = 0;
			annual_dt = 0;
			misfit = 0;
			time = starttime;
			femmodel->parameters->SetParam(starttime,TimeEnum);
			femmodel->parameters->SetParam(step,StepEnum);
			while(time < finaltime + (yts*DBL_EPSILON)){ //make sure we run up to finaltime.
				/*Time Increment*/
				switch(timestepping){
					case AdaptiveTimesteppingEnum:
						femmodel->TimeAdaptx(&dt);
						if(time+dt>finaltime) dt=finaltime-time;
						femmodel->parameters->SetParam(dt,TimesteppingTimeStepEnum);
						break;
					case FixedTimesteppingEnum:
						femmodel->parameters->FindParam(&dt,TimesteppingTimeStepEnum);
						break;
					default:
						_error_("Time stepping \""<<EnumToStringx(timestepping)<<"\" not supported yet");
				}
				step+=1;
				time+=dt;
				annual_dt+=dt;
				femmodel->parameters->SetParam(time,TimeEnum);
				femmodel->parameters->SetParam(step,StepEnum);

				_printf0_(" MisfitAnnual::197 accumulating...\n");
				//Accumulate mean misfit on each element
				for(Object* & object : femmodel->elements->objects){
					Element* element=xDynamicCast<Element*>(object);
					element->MisfitAccumulate(model_enum,dt);
				}

				if (annual_dt >= 0) {
				//if (annual_dt >= (yts - yts*DBL_EPSILON)) {
					//Return mean misfit on each element
					misfit_t = 0;
					area_t = 0;

					for(Object* & object : femmodel->elements->objects){
						Element* element=xDynamicCast<Element*>(object);
						area_e=element->MisfitArea(weights_enum);
						misfit_t+=element->MisfitAnnual(model_enum,observation_enum,weights_enum,annual_dt);
						area_t+=area_e;
					}
					ISSM_MPI_Allreduce ( (void*)&misfit_t,(void*)&all_misfit_t,1,ISSM_MPI_DOUBLE,ISSM_MPI_SUM,IssmComm::GetComm());
					ISSM_MPI_Allreduce ( (void*)&area_t,(void*)&all_area_t,1,ISSM_MPI_DOUBLE,ISSM_MPI_SUM,IssmComm::GetComm());
					area_t=all_area_t;
					misfit_t=all_misfit_t;

					/*Divide by surface area if not nill!: */
					if (area_t!=0) misfit_t=misfit_t/area_t;

					/*Add this time's contribution to curent misfit: */
					misfit+=misfit_t/annual_dt/(finaltime-starttime);
					_printf0_(" MisfitAnnual::216 step: " << step << " misfit: " << misfit << " misfit_contrib:" << misfit_t*annual_dt/(finaltime-starttime) << "\n");
					annual_dt = 0;
				}
			}
			_printf0_(" MisfitAnnual::227 returning misfit: " << misfit << "\n");
			return misfit;
		}
		if(time==finaltime)this->lock+=1;              
		return misfit;
    } /*}}}*/
    else if (this->local==2){ /*vertex by vertex computation: {{{*/

   	 IssmDouble* model = NULL;
   	 IssmDouble* observation= NULL;
   	 IssmDouble* weights= NULL;
   	 int msize,osize,wsize;

   	 /*Are we transient?:*/
   	 if (time==0){
   		 IssmDouble misfit_t=0.;

   		 /*get global vectors: */
   		 GetVectorFromInputsx(&model,&msize,femmodel,model_enum);
   		 GetVectorFromInputsx(&observation,&osize,femmodel,observation_enum);_assert_(msize==osize);
   		 GetVectorFromInputsx(&weights,&wsize,femmodel,weights_enum); _assert_(wsize==msize);

   		 int count=0;
   		 for (int i=0;i<msize;i++){
   			 misfit_t += pow(model[i]-observation[i],2)*weights[i];
   			 if (weights[i]!=0)count++;
   		 }
   		 misfit=sqrt(misfit_t/count);

   		 /*Free resources:*/
   		 xDelete<IssmDouble>(model);
   		 xDelete<IssmDouble>(observation);
   		 xDelete<IssmDouble>(weights);

   		 /*return value: */
   		 return misfit;
   	 }
   	 else{

   		 IssmDouble misfit_t=0.;
   		 IssmDouble all_misfit_t=0.;

   		 /*If we are locked, return time average: */
   		 if(this->lock)return misfit/(time-starttime);

   		 /*get global vectors: */
   		 GetVectorFromInputsx(&model,&msize,femmodel,model_enum);
   		 GetVectorFromInputsx(&observation,&osize,femmodel,observation_enum);_assert_(msize==osize);
   		 GetVectorFromInputsx(&weights,&wsize,femmodel,weights_enum); _assert_(wsize==msize);

   		 int count=0;
   		 for (int i=0;i<msize;i++){
   			 misfit_t += pow(model[i]-observation[i],2)*weights[i];
   			 if (weights[i]!=0)count++;
   		 }
   		 misfit=sqrt(misfit_t/count);

   		 /*Add this time's contribution to curent misfit: */
   		 misfit=sqrt(misfit_t)/count;
   		 misfit+=dt*misfit_t;

   		 /*Do we lock? i.e. are we at final_time? :*/
   		 if(time==finaltime)this->lock=1;

   		 /*Free resources:*/
   		 xDelete<IssmDouble>(model);
   		 xDelete<IssmDouble>(observation);
   		 xDelete<IssmDouble>(weights);

   		 /*What we return is the value of misfit / time: */
   		 return misfit/(time-starttime);
   	 }

    } /*}}}*/
    else{ /*global computation: {{{ */

   		IssmDouble model, observation;

   		/*If we are locked, return time average: */
   		if(this->lock) return misfit/(time-starttime);

   		/*First, the global  model response: */
   		model=OutputDefinitionsResponsex(femmodel,this->model_enum);
   		/*Now, the observation is buried inside the elements, go fish it in the first element (cludgy, needs fixing): */
   		Element* element = (Element*)femmodel->elements->GetObjectByOffset(0); _assert_(element);
   		Input*  input   = element->GetInput(observation_enum); _assert_(input);
   		input->GetInputAverage(&observation);

   		/*Add this time's contribution to curent misfit: */
   		misfit+=dt*(model-observation);

   		/*Do we lock? i.e. are we at final_time? :*/
   		if(time==finaltime)this->lock=1;

   		/*What we return is the value of misfit / time: */
   		return misfit/(time-starttime);
    } /*}}}*/

}
	/*}}}*/
